from __future__ import annotations

from collections.abc import Callable
from types import TracebackType

from sqlalchemy.orm.session import Session as SASession

from .transaction import Transaction


class Session:
    """Wrapper for SQLAlchemy session objects.

    Initialize by providing with a `sessionmaker` object.

    Must be used as context manager.
    """

    def __init__(self, session_maker: Callable[[], SASession]) -> None:
        self._session_maker = session_maker
        self._session: SASession | None = None
        self._transaction: Transaction | None = None

    def __enter__(self) -> Session:
        if self._session:
            raise RuntimeError("session already entered")
        self._session = self._session_maker()
        assert self._session
        self._transaction = Transaction(self._session)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if not self._session:
            raise RuntimeError("not in a session context")
        if exc_type:
            self._session.rollback()
        else:
            self._session.commit()
        self._session.close()
        self._session = None
        self._transaction = None

    @property
    def transaction(self) -> Transaction:
        """Access the transaction directly."""
        if self._transaction is None:
            raise RuntimeError("not in a session context")
        return self._transaction

    def begin_transaction(self) -> Transaction:
        """Start a new transaction.

        Use as a context manager:

        >>> with session.begin_transaction() as t:
        ...     ...

        It is not possible to nest calls to begin_transaction().
        """
        if not self._transaction:
            raise RuntimeError("not in a session context")
        return self._transaction



from .base import DBObjectBase as DBObjectBase  # noqa F401
from .exc import (  # noqa F401
    DataError as DataError,
    UnknownItemError as UnknownItemError,
)
from .session import Session as Session  # noqa F401
from .transaction import Transaction as Transaction  # noqa F401



from __future__ import annotations

from typing import Any


class DataError(Exception):
    """Base class for sqla-utils exceptions."""


class UnknownItemError(DataError):
    """An unknown item was queried.

    By default, a message a is generated from the supplied attributes, but
    a custom message can be provided.

    Arguments and fields:
    item_type -- the kind item that failed to query, often the table name
    field -- the field the query failed on, often a column name
    value -- the value queried
    """

    def __init__(
        self,
        item_type: str,
        field: str | None = None,
        value: Any = None,
        *,
        msg: str | None = None,
    ) -> None:
        if msg is None:
            msg = f"unknown '{item_type}' item"
            if field:
                msg += f", no {field} with value '{value!r}'"
        super().__init__(msg)
        self.item_type = item_type
        self.field = field
        self.value = value



from __future__ import annotations

import re
from os import PathLike
from pathlib import Path
from typing import Any, Callable, Iterable

from .split_sql import split_sql

SQLExecutor = Callable[[str], Any]


class DependencyLoopError(Exception):
    pass


class DatabaseBuilder:
    """Automatic SQL database builder.

    Apply select SQL scripts from a given directory to the given SQL engine.

    SQL scripts can require other SQL scripts to be read beforehand.

    >>> class MyEngine:
    ...     def execute(self, query):
    ...         print(query)
    ...
    >>> engine = MyEngine()
    >>> f = open("feature1.sql", "w")
    >>> f.write("-- Require: feature2\\n\\nSELECT * FROM feature1;")
    >>> f.close()
    >>> f = open("feature2.sql", "w")
    >>> f.write("SELECT * FROM feature2;")
    >>> f.close()
    >>> builder = DatabaseBuilder(engine, ".")
    >>> builder.require("feature1")
    SELECT * FROM feature2
    SELECT * FROM feature1
    >>> os.remove("feature1.sql")
    >>> os.remove("feature2.sql")
    >>>
    """

    def __init__(self, executor: SQLExecutor, path: PathLike | str) -> None:
        self._executor = executor
        self._path = Path(path)
        self._parsed: set[str] = set()
        self._parsing: list[str] = []

    def require(self, *requirements: str) -> None:
        for requirement in requirements:
            if requirement not in self._parsed:
                if requirement in self._parsing:
                    raise DependencyLoopError(
                        "dependency loop: " + requirement
                    )
                self._require_one(requirement)
                self._parsed.add(requirement)

    def _require_one(self, requirement: str) -> None:
        self._parsing.append(requirement)
        try:
            req_file = self._path / (requirement + ".sql")
            with open(req_file) as f:
                headers = _parse_sql_headers(f)
                self._add_requires(headers.get("require", ""))
            with open(req_file) as f:
                _execute_sql_stream(self._executor, f)
        finally:
            self._parsing.pop()

    def _add_requires(self, requires_string: str) -> None:
        if requires_string.strip():
            requires = [r.strip() for r in requires_string.split(",")]
            self.require(*requires)


_SQL_LINE_RE = re.compile(
    r"^--+\s+((?:[a-zA-Z][a-zA-Z0-9]*)(?:-[a-zA-Z][a-zA-Z0-9]*)*):\s+(.*)$"
)


def _parse_sql_headers(stream: Iterable[str]) -> dict[str, str]:
    matches = []
    for line in stream:
        m = _SQL_LINE_RE.match(line)
        if not m:
            break
        matches.append(m)
    return {m.group(1).lower(): m.group(2).strip() for m in matches}


def _execute_sql_stream(executor: SQLExecutor, stream: Iterable[str]) -> None:
    """Run the SQL statements in a stream against a database."""
    for query in split_sql(stream):
        query = query.replace("%", "%%")
        executor(query)



from __future__ import annotations

import re
from typing import Generator, Iterable, Iterator

_SQL_COMMENT_START = "--"
_SQL_STRING_DELIMITER = "'"
_SQL_STATEMENT_DELIMITER = ";"

_DELIMITER_LINE_RE = re.compile(r"^\s*delimiter\s+(.*)\s*$", re.IGNORECASE)


def split_sql(stream: Iterable[str]) -> Iterator[str]:
    """Return an iterator over the SQL statements in a stream.

    >>> list(split_sql("SELECT * FROM foo; DELETE FROM foo;"))
    ... ['SELECT * FROM foo', 'DELETE FROM foo']
    """
    return _SQLSplitter(stream).split()


class _SQLSplitter:
    def __init__(self, stream: Iterable[str]) -> None:
        self._stream = stream
        self._current_stmt = ""
        self._in_string = False
        self._stmt_delimiter = _SQL_STATEMENT_DELIMITER
        self._line = ""

    def split(self) -> Iterator[str]:
        for self._line in self._stream:
            yield from self._parse_line()
        if self._current_stmt.strip():
            yield self._current_stmt.strip()

    def _parse_line(self) -> Generator[str, None, None]:
        m = _DELIMITER_LINE_RE.match(self._line)
        if m:
            self._stmt_delimiter = m.group(1)
        else:
            yield from self._parse_sql_line()

    def _parse_sql_line(self) -> Generator[str, None, None]:
        self._pos = 0
        while self._pos < len(self._line):
            if self._at_comment_start():
                break
            elif self._at_statement_delimiter():
                if self._current_stmt.strip():
                    yield self._current_stmt.strip()
                self._current_stmt = ""
                self._pos += len(self._stmt_delimiter)
            elif self._at_string_delimiter():
                self._in_string = not self._in_string
                self._current_stmt += _SQL_STRING_DELIMITER
                self._pos += len(_SQL_STRING_DELIMITER)
            else:
                self._current_stmt += self._line[self._pos]
                self._pos += 1

    def _at_comment_start(self) -> bool:
        return self._current_pos_startswith(_SQL_COMMENT_START)

    def _at_statement_delimiter(self) -> bool:
        if self._in_string:
            return False
        return self._current_pos_startswith(self._stmt_delimiter)

    def _at_string_delimiter(self) -> bool:
        return self._current_pos_startswith(_SQL_STRING_DELIMITER)

    def _current_pos_startswith(self, s: str) -> bool:
        return self._line[self._pos :].startswith(s)



from __future__ import annotations

from typing import Any, TypeVar
from uuid import UUID

from sqlalchemy import Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Query

from .exc import UnknownItemError
from .transaction import Transaction

_DB = TypeVar("_DB", bound="DBObjectBase")

_DeclarativeBase = declarative_base()


class DBObjectBase(_DeclarativeBase):  # type: ignore
    """Base class for mapped classes.

    Each non-abstract sub-class must provide a __tablename__
    or a __table__ class attribute.
    """

    __abstract__ = True
    __tablename__: str
    __table__: Table

    @classmethod
    def item_type(cls) -> str:
        """Name of the kind of items this represents.

        Used for describing exceptions. Can be overriden by sub-classes.
        Defaults to the table name.
        """
        if hasattr(cls, "__tablename__"):
            return cls.__tablename__
        elif hasattr(cls, "__table__"):
            return cls.__table__.name
        else:
            raise RuntimeError(f"{cls!r} missing __table__ and __tablename__")

    @classmethod
    def query(
        cls: type[_DB],
        t: Transaction,
        *conditions: Any,
        order_by: Any | None = None,
    ) -> Query[_DB]:
        """Return an SQLAlchemy query for a list of database entries.

        Return a query for all entries of this class's table by default,
        but the query can be narrowed by supplying filter conditions.
        """
        query = t.query(cls).filter(*conditions)
        if order_by is not None:
            query = query.order_by(order_by)
        return query

    @classmethod
    def count(cls, t: Transaction, *conditions: Any) -> int:
        """Return number of entries for a given query condition.

        Return the total number of entries if no filter condition
        is specified.
        """
        return cls.query(t, *conditions).count()

    @classmethod
    def first(
        cls: type[_DB],
        query: Query[_DB],
        *,
        field: str | None = None,
        value: Any = None,
    ) -> _DB:
        """Return a single entry from a query.

        If the database contains multiple entries that match the given
        conditions, return an arbitrary entry. If it contains no matching
        entries, raise a UnknownItemError.
        """
        o: _DB | None = query.first()
        if o is None:
            raise UnknownItemError(cls.item_type(), field, value)
        return o

    @classmethod
    def fetch_all(
        cls: type[_DB],
        t: Transaction,
        *conditions: Any,
        order_by: Any | None = None,
    ) -> list[_DB]:
        """Return an a list of entries from the database.

        Return all entries of this class's table by default, but the
        query can be narrowed by supplying filter conditions.
        """
        return cls.query(t, *conditions, order_by=order_by).all()

    @classmethod
    def fetch_one(
        cls: type[_DB],
        t: Transaction,
        *conditions: Any,
        order_by: Any | None = None,
        field: str | None = None,
        value: Any = None,
    ) -> _DB:
        """Return a single entry from the database.

        If the database contains no matching entries, raise an
        UnknownItemError. If it contains multiple entries that match the given
        conditions, return the first entry according to the given order
        condition. If no condition is given, return an arbitrary entry.
        """
        q = cls.query(t, *conditions, order_by=order_by)
        return cls.first(q, field=field, value=value)

    @classmethod
    def fetch_by_id(cls: type[_DB], t: Transaction, id: int) -> _DB:
        """Return the database entry with the given id.

        For this method to work, the database table needs to have a
        numeric column named "id".
        """
        return cls.fetch_one(t, cls.id == id, field="id", value=id)

    @classmethod
    def fetch_by_tag(cls: type[_DB], t: Transaction, tag: str) -> _DB:
        """Return the database entry with the given tag.

        For this method to work, the database table needs to have a
        string "tag" column.
        """
        return cls.fetch_one(t, cls.tag == tag, field="tag", value=tag)

    @classmethod
    def fetch_by_uuid(cls: type[_DB], t: Transaction, uuid: UUID) -> _DB:
        """Return the database entry with the given UUID.

        For this method to work, the database table needs to have a
        string "uuid" column.
        """
        return cls.fetch_one(
            t, cls._uuid == str(uuid), field="uuid", value=uuid
        )

    @classmethod
    def delete_all(cls, t: Transaction, *conditions: Any) -> None:
        """Delete all entries that match certain conditions.

        If no conditions are given, delete all entries from the table.
        """
        cls.query(t, *conditions).delete()

    @classmethod
    def delete_one(cls, t: Transaction, *conditions: Any) -> None:
        """Delete an entry that matches certain conditions.

        If the database contains multiple entries that match the given
        conditions, delete an arbitrary entry. If it contains no matching
        entries, raise a UnknownItemError.
        """
        o = cls.fetch_one(t, *conditions)
        t.delete(o)

    @classmethod
    def delete_by_id(cls, t: Transaction, id: int) -> None:
        """Delete the entry with the given id.

        Raise UnknownItemError if no entry with this id exists.

        For this method to work, the database table needs to have a
        numeric column named "id".
        """
        item = cls.fetch_by_id(t, id)
        item.delete(t)

    @classmethod
    def delete_by_tag(cls, t: Transaction, tag: str) -> None:
        """Delete the entry with the given tag.

        Raise UnknownItemError if no entry with this tag exists.

        For this method to work, the database table needs to have a
        string column named "tag".
        """
        item = cls.fetch_by_tag(t, tag)
        item.delete(t)

    @classmethod
    def delete_by_uuid(cls, t: Transaction, uuid: UUID) -> None:
        """Delete the entry with the given UUID.

        Raise UnknownItemError if no entry with this UUID exists.

        For this method to work, the database table needs to have a
        numeric column named "uuid".
        """
        item = cls.fetch_by_uuid(t, uuid)
        item.delete(t)

    def delete(self, t: Transaction) -> None:
        """Delete this entry from the database."""
        t.delete(self)



"""pytest utilities for working with SQLAlchemy."""

from __future__ import annotations

import os
import shutil
from pathlib import Path
from tempfile import mkstemp
from types import TracebackType
from typing import Any, Iterable, Mapping, Sequence, TypeVar

import pytest
from sqlalchemy.engine import Connection, create_engine, Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import insert, select
from sqlalchemy.sql.schema import MetaData, Table

from .builder import DatabaseBuilder
from .session import Session
from .types import RowType

_S = TypeVar("_S", bound="DBFixture")

_MEMORY_DB_URL = "sqlite:///:memory:"


def assert_row_equals(
    row: Mapping[str, Any], expected_values: Mapping[str, Any]
) -> None:
    """Assert that a row contains expected values.

    row is a mapping as returned from execute_sql_one_row() or
    fetch_only_row(). expected_values is a column name -> expected value
    mapping. The row may contain additional columns that are not listed
    in expected_values. Those are ignored.
    """

    for column_name, expected in expected_values.items():
        column_value = row[column_name]
        assert (
            column_value == expected
        ), f"column '{column_name}': {expected!r} != {column_value!r}"


def assert_one_row_equals(
    rows: Iterable[Mapping[str, Any]], expected_values: Mapping[str, Any]
) -> None:
    """Assert that one of a list of rows contains the expected values.

    rows is a list of mappings as returned from execute_sql() or
    fetch_all_rows(). expected_values is a column name -> expected value
    mapping. The row may contain additional columns that are not listed
    in expected_values. Those are ignored.
    """

    def check_row(row: Mapping[str, Any]) -> bool:
        try:
            assert_row_equals(row, expected_values)
        except AssertionError:
            return False
        else:
            return True

    if not any(check_row(r) for r in rows):
        pytest.fail("no row is matching the expectation")


class DBFixture:
    """Database test fixture for pytest and SQLalchemy.

    This uses SQLite to test the SUT, which means that it
    can't be used to test database-specific code.

    Databases fixtures can be set up with one of two methods. Either, by
    supplying a path to a directory containing SQL files that configure
    the database. Example:

        >>> class MyFixture(DBFixture):
        ...     __metadata__ = DBObjectBase.metadata
        ...     sql_path = Path("./foo")
        ...     requirements = ["items"]

    Alternatively, it can be pointed to a template SQLite database that
    will be copied. Example:

        >>> class MyFixture(DBFixture):
        ...     __metadata__ = DBObjectBase.metadata
        ...     db_path = Path("./foo.sqlite")

    The fixture sub-class should contain utility methods for setting up
    and querying the database:

        >>> class MyFixture(DBFixture):
        ...     __metadata__ = DBObjectBase.metadata
        ...     db_path = Path("./foo.sqlite")
        ...
        ...     def insert_item(self, *, id: int, text: str) -> int:
        ...         self.insert("items", {"id": id, "text": text})
        ...         return id
    """

    __metadata__: MetaData = MetaData()
    sql_path: Path | None = None
    db_path: Path | None = None
    requirements: list[str] = []

    def __init__(self) -> None:
        if self.sql_path is None and self.db_path is None:
            raise RuntimeError("either of sql_path or db_path must be set")
        if self.sql_path is not None and self.db_path is not None:
            raise RuntimeError("only one of sql_path and db_path can be set")
        self._tmp_db: Path | None = None
        self.engine: Engine | None = None
        self._connection: Connection | None = None
        self._db_builder: DatabaseBuilder | None = None
        self._session: Session | None = None

    def __enter__(self: _S) -> _S:
        if self.db_path is not None:
            self._tmp_db = _copy_database(self.db_path)
        self.engine = create_engine(self.db_url)
        self._connection = self.engine.connect()
        self._connection.execute("PRAGMA foreign_keys=ON")
        self.__metadata__.bind = self.engine
        self._session = Session(sessionmaker(bind=self.engine)).__enter__()

        if self.sql_path is not None:
            self._db_builder = DatabaseBuilder(
                self.connection.execute, str(self.sql_path)
            )
        if self.requirements:
            self.require(*self.requirements)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._session is not None
        assert self._connection is not None
        self._session.__exit__(exc_type, exc_val, exc_tb)
        self._session = None
        self._connection.close()
        self._connection = None
        self.engine = None
        if self._tmp_db is not None:
            os.remove(self._tmp_db)
            self._tmp_db = None

    @property
    def db_url(self) -> str:
        if self._tmp_db is not None:
            return f"sqlite:///{self._tmp_db}"
        else:
            return _MEMORY_DB_URL

    @property
    def connection(self) -> Connection:
        if self._connection is None:
            raise RuntimeError("call __enter__() before accessing connection")
        return self._connection

    @property
    def session(self) -> Session:
        if self._session is None:
            raise RuntimeError("call __enter__() before accessing session")
        return self._session

    def require(self, *features: str) -> None:
        if self._db_builder is None:
            raise RuntimeError("SQL database not built dynamically")
        self._db_builder.require(*features)

    def execute_sql(
        self, query: str, args: Sequence[Any] | None = None
    ) -> None:
        """Execute a SQL query."""
        if args is None:
            res = self.connection.execute(query)
        else:
            res = self.connection.execute(query, args)
        res.close()

    def select_sql(
        self, query: str, args: Sequence[Any] | None = None
    ) -> list[RowType]:
        """Execute a SQL SELECT and return all rows."""
        if args is None:
            res = self.connection.execute(query)
        else:
            res = self.connection.execute(query, args)
        try:
            return res.fetchall()
        finally:
            res.close()

    def select_sql_one_row(
        self, query: str, args: Sequence[Any] | None = None
    ) -> RowType:
        """Execute a SQL SELECT and return one row.

        Raise an AssertionError if the result has zero or more than one row.
        """
        rows = self.select_sql(query, args)
        assert len(rows) == 1, f"got {len(rows)} rows, expected 1"
        return rows[0]

    def select_all_rows(self, table_name: str) -> list[RowType]:
        """Return all rows from a table."""
        table = Table(table_name, self.__metadata__, autoload=True)
        return self.select_sql(select([table]))

    def select_only_row(self, table_name: str) -> RowType:
        """Return the only row from a table.

        Raise an AssertionError if the table has zero or more than one row."""
        rows = self.select_all_rows(table_name)
        assert len(rows) == 1, (
            f"expected exactly one row in table '{table_name}', "
            "got {len(rows)}"
        )
        return rows[0]

    def insert(self, table_name: str, values: Mapping[str, Any]) -> None:
        """Insert one row into a table using a mapping of values."""
        table = Table(table_name, self.__metadata__, autoload=True)
        self.connection.execute(insert(table, values))

    def assert_table_is_empty(self, table_name: str) -> None:
        """Assert that a table has no rows."""
        rows = self.select_all_rows(table_name)
        assert len(rows) == 0, (
            f"table {table_name} contains {len(rows)} rows, "
            "expected it to be empty"
        )

    def assert_row_count(self, table_name: str, expected_rows: int) -> None:
        """Assert that a table has a certain amount of rows."""
        rows = self.select_all_rows(table_name)
        assert len(rows) == expected_rows, (
            f"table {table_name} contains {len(rows)} rows, "
            f"expected {expected_rows}"
        )

    def assert_only_row_equals(
        self, table_name: str, expected_values: Mapping[str, Any]
    ) -> None:
        """Assert that a table contains only one row with expected values.

        expected_values is a column name -> expected value
        dictionary. The row may contain additional columns that are not listed
        in expected_values. Those are ignored.
        """
        row = self.select_only_row(table_name)
        assert_row_equals(row, expected_values)

    def assert_any_row_equals(
        self, table_name: str, expected_values: Mapping[str, Any]
    ) -> None:
        """Assert that a table contains a row with expected values.

        expected_values is a column name -> expected value
        dictionary. The row may contain additional columns that are not listed
        in expected_values. Those are ignored.
        """
        for row in self.select_all_rows(table_name):
            try:
                assert_row_equals(row, expected_values)
            except AssertionError:
                pass
            else:
                return
        pytest.fail("no row matches the expectations")

    def assert_rows_equal(
        self, table_name: str, expected_rows: Sequence[Mapping[str, Any]]
    ) -> None:
        """Assert that the expected rows are in table and no other rows."""
        __tracebackhide__ = True

        fetched_rows = self.select_all_rows(table_name)
        if not fetched_rows and not expected_rows:
            return
        assert len(fetched_rows) == len(expected_rows)

        def find_one(
            rs: list[RowType], expected: Mapping[str, Any]
        ) -> list[RowType]:
            __tracebackhide__ = True
            for i, tr in enumerate(rs):
                try:
                    assert_row_equals(tr, expected)
                except AssertionError:
                    pass
                else:
                    return rs[:i] + rs[i + 1 :]
            pytest.fail(f"no row matching {expected} found")

        for row in expected_rows:
            fetched_rows = find_one(fetched_rows, row)


def _copy_database(path: Path) -> Path:
    fd, name = mkstemp(".sqlite", "test-")
    dst = os.fdopen(fd, "wb")
    src = open(path, "rb")
    shutil.copyfileobj(src, dst)
    src.close()
    dst.close()
    return Path(name)



from __future__ import annotations

from types import TracebackType
from typing import Any, Iterable, Mapping, TypeVar, overload

from sqlalchemy.engine import Connection
from sqlalchemy.orm import Query, Session
from sqlalchemy.schema import Table
from sqlalchemy.sql import ColumnElement

_T = TypeVar("_T")
_TA = TypeVar("_TA", bound="Transaction")


class Transaction:
    """Wrapper around SQLAlchemy sessions and transactions.

    Can be used as a context manager to create a new transaction
    that will be committed or rollbacked when the context is
    exited.

    >>> with Transaction(...) as t:
    ...     ...
    """

    def __init__(self, session: Session) -> None:
        self.session = session

    @property
    def connection(self) -> Connection:
        return self.session.connection()

    def __enter__(self: _TA) -> _TA:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        try:
            self.session.flush()
        except BaseException:
            self.session.rollback()
            raise
        if exc_type:
            self.session.rollback()
        else:
            self.session.commit()

    @overload
    def query(self, entities: Table, **kwargs: Any) -> Query[Any]:
        ...

    @overload  # noqa: F811
    def query(
        self, *entities: type[_T], **kwargs: Any
    ) -> Query[_T]:  # noqa: F811
        ...

    @overload  # noqa: F811
    def query(  # type: ignore  # noqa: F811
        self, entities: ColumnElement[_T], **kwargs: Any
    ) -> Query[tuple[_T]]:
        ...

    @overload  # noqa: F811
    def query(  # noqa: F811
        self, *entities: ColumnElement[_T], **kwargs: Any
    ) -> Query[tuple[_T, ...]]:
        ...

    def query(self, *entities: Any, **kwargs: Any) -> Any:  # noqa: F811
        """Wrapper around Session.query()."""
        return self.session.query(*entities, **kwargs)

    def add(self, *instances: Any) -> None:
        """Save one or more objects to the database."""
        self.session.add_all(instances)
        self.flush(*instances)

    def delete(self, *instances: Any) -> None:
        """Mark one or more instances as deleted."""
        for obj in instances:
            self.session.delete(obj)
        self.flush(*instances)

    def flush(self, *objects: Any) -> None:
        """Flush object changes to the database.

        As opposed to Session.flush() this takes the objects
        to flush as positional arguments. Flush all changes
        if no objects are provided.
        """
        if len(objects) == 0:
            self.session.flush()
        else:
            self.session.flush(objects)

    def refresh(self, *instances: Any) -> None:
        """Wrapper around Session.refresh.

        Can be called with multiple instances.
        """
        for instance in instances:
            self.session.refresh(instance)

    def expire_all(self) -> None:
        """Wrapper around Session.expire_all()."""
        self.session.expire_all()

    def execute(
        self,
        query: Any,
        args: Mapping[str, Any] | Iterable[Mapping[str, Any]] | None = None,
    ) -> Any:
        """Wrapper around Session.execute()."""
        return self.session.execute(query, args)

    def scalar(
        self,
        query: Any,
        params: Mapping[str, Any] | Iterable[Mapping[str, Any]] | None = None,
    ) -> Any:
        """Wrapper around Session.scalar()."""
        return self.session.scalar(query, params)
